package protocol;

import java.util.Random;
import file_system.*;
import multicast.*;
import constants.Constants;
import exceptions.*;

public class Chunk_dispatcher_1 {
	private String packet;
	private boolean temp_1;

	public Chunk_dispatcher_1(String packet)
	{
		this.packet = packet;
		temp_1 = false;
	}

	public Chunk_dispatcher_1(String packet, boolean temp_2)
	{
		this.packet = packet;
		temp_1 = temp_2;
	}

	public void put_chunk(String file_id, int chunk_number, String data, int duplication_degree)
	{                              
		Chunk chunk = Save_chunk.get_current_chunk(file_id, chunk_number);
		if(chunk == null)
		{
			chunk = Stored_files.get_stored_files().get_file(file_id).get_chunk(chunk_number);
			if(chunk == null)
			{
				chunk = new Chunk(file_id,chunk_number,data,duplication_degree);
			}
			Save_chunk.current_chunks.add(chunk);
		}
		Reclaim_space_1.check_chunk(chunk);
		chunk.start_duplication_degree();
		Random random = new Random();
		synchronized(this)
		{
			try {
				wait( random.nextInt(400));
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		if( duplication_degree > chunk.get_current_duplication_degree())
		{
			Save_chunk.send_chunk(file_id, chunk_number);        
			chunk.store();
			Stored_files.get_stored_files().get_file(file_id).add_chunk(chunk);
			Stored_files.save_chunk(chunk);
		}
		else
		{
			Stored_files.get_stored_files().get_file(file_id).delete_chunk(chunk_number);
			Stored_files.delete_chunk(chunk);
		}
		Save_chunk.delete_current_chunk(file_id, chunk_number);
	}

	private void store_chunk(int chunk_number, String file_id) throws No_such_chunk
	{
		Chunk chunk = Save_chunk.get_current_chunk(file_id, chunk_number);
		if( chunk == null )
		{
			chunk = Stored_files.get_stored_files().get_file(file_id).get_chunk(chunk_number);
		}
		if( chunk == null )
		{
			throw new No_such_chunk();
		}
		chunk.store();
	}

	private void send_file(String file_id,int duplication_degree) {
		Files file = Stored_files.get_stored_files().get_file(file_id);
		synchronized(file)
		{
			int chunk_number = 0;
			for( int i=0; i<file.size(); i++)
			{
				Chunk chunk = file.get_chunk(chunk_number);
				while(chunk==null)
				{
					chunk_number++;
					chunk = file.get_chunk(chunk_number);
				}
				Save_chunk.current_chunks.add(chunk);
				send_put(file_id,chunk_number,duplication_degree);
				Save_chunk.current_chunks.remove(chunk);
				chunk_number++;
			}
		}
		System.out.println("File " + file_id + "has been sent.");
	}

	private void send_put(String file_id, int chunk_number, int duplication_degree)  {            
		Chunk chunk = Save_chunk.get_current_chunk(file_id, chunk_number);
		packet = Save_chunk.get_putchunk_packet(chunk);
		chunk.start_duplication_degree();
		int waiting_time = 500, i=0;
		while(i < 5 && duplication_degree > chunk.get_current_duplication_degree())
		{
			chunk.start_duplication_degree();
			Stored_data_1.MDB().send(packet);
			synchronized(this)
			{
				try {
					wait(waiting_time);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			waiting_time = waiting_time*2;
			i++;
		}
		Save_chunk.delete_current_chunk(file_id, chunk_number);
	}

	public void run()
	{
		String packet_temp[] = packet.split(Constants.CRLF());
		String head = packet_temp[0];
		String head_temp[] = head.split(" ");
		String file_id = head_temp[2],  messageType = head_temp[0];
		int chunk_number = Integer.parseInt(head_temp[3]);
		if( messageType.equalsIgnoreCase("PUTCHUNK") && !(temp_1)){
			String data;
			if( packet_temp.length < 3 )
			{
				data = "";
			}
			else
			{
				data = packet_temp[2];                
			}
			data = data.trim();
			int duplication_degree = Integer.parseInt(head_temp[4]);
			put_chunk(file_id, chunk_number, data, duplication_degree);
		}
		else if( messageType.equalsIgnoreCase("PUTCHUNK") && temp_1 )
		{
			int duplication_degree = Integer.parseInt(head_temp[4]);
			send_put(file_id,chunk_number,duplication_degree);

		}
		else if( messageType.equalsIgnoreCase("PUTFILE") && temp_1 )
		{
			int duplication_degree = Integer.parseInt(head_temp[4]);
			send_file(file_id,duplication_degree);
		}
		else if( messageType.equalsIgnoreCase("STORED") && !temp_1 )
		{
			try {
				store_chunk(chunk_number,file_id);
			} catch (No_such_chunk e) {

			}
		}
	}
}